""" package implementing useful rayoptics graphics using matplotlib

    The ``rayoptics.mpl`` subpackage provides useful basic optical graphics
    using the matplotlib plotting package. Particular features include:

        - 2D lens layout
        - ray aberration and wavefront pupil/field plots
        - |ybar| and |nubar| paraxial ray diagrams
"""
