""" Package encompassing platform agnostic gui commands and updating

    The ``rayoptics.gui`` subpackage provides a thin app manager framework
    :mod:`~.appmanager` and functions implementing basic commands.

    The :mod:`~.layout` module provides support for interactive editing of
    optical layouts. This includes support for selecting and drawing different
    graphical parts of the optical model.
"""
