""" package supplying utility functions for math and numpy support

    The ``rayoptics.util`` subpackage provides miscellaneous functions for
    geometric calculations, color calculations and anything else that doesn't
    have an obvious home.
"""
