""" package supplying Qt5 desktop application and associated functional support

    The ``rayoptics.qtgui`` subpackage provides a desktop app that runs under
    Anaconda. It also provides a series of higher level interfaces used by
    rayoptics. These include:

        - an interface that hosts matplotlib graphics
        - a table grid for numeric model displays (template-based)
        - docking panel support for python objects
        - iPython console window (desktop app only)
"""
